#include <assert.h>
#include <err.h>
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>

#include <SDL/SDL.h>
#include <SDL/SDL_audio.h>
#include <SDL/SDL_timer.h>

#include <sys/time.h>

const int       AMPLITUDE = 7000;
const int       FREQUENCY = 44100;

int
min(int a, int b)
{
	if (a < b)
	{
		return a;
	}
	return b;
}

/*
 * This carries all the state between sounds
 */
struct userdata
{
	int            *counters;
	int            *primes;
	int            *freq;
	int            *amp;
	int             c;
	int             n;
	double         *v;
	int             repeats;/* unused */
};


void
init_primes(int *ar, int n)
{
	assert(n > 2);
	int             i, j, p, primep;
	ar[0] = 2;
	p = 1;
	for (i = 1; i < n; i++)
	{
		primep = 0;
hell:
		p = p + 2;
		for (j = 0; j < i; j++)
		{
			if (p % ar[j] == 0)
			{
				goto hell;
				break;
			}
		}
		ar[i] = p;
	}
	return;
}

/*
 * Does the math
 */
void
bstep(struct userdata * u)
{
	int            *cs = u->counters;
	int            *ps = u->primes;
	int             n = u->n;
	int            *hz = u->freq;
	int            *amp = u->amp;
	int             i, c = 0;
	for (i = 0; i < n; i++)
	{
		cs[i] = (cs[i] + 1) % ps[i];
		if (cs[i] == 0)
		{
			hz[c] = (i * (2000 / u->n)) + 40;	/* scale it to audible
								 * sound (heuristically) */
			amp[c] = AMPLITUDE * log2(ps[i]);
			hz[++c] = -1;	/* end of list marker  (stupid C) */
		}
	}
	if (u->c % 1000 == 0)
		fprintf(stderr, "%d\n", u->c);
	u->c++;
}


/*
 * This is called for every sample that SDL is willing to play,
 */
void
audio_callback(void *ud, Uint8 * stream, int length)
{
	int             j, i = 0;
	int             l = length / 2;	/* shorts, not chars */
	int             nf;
	Sint16         *s = (Sint16 *) stream;
	struct userdata *u = ud;

	for (j = 0; u->freq[j] != -1; j++)
	{
	}
	nf = j;

	while (i < l)
	{
		for (j = 0; j < nf; j++)
		{
			s[i] += u->amp[j] * sin(u->v[j] * 2 * M_PI / FREQUENCY);
			u->v[j] += u->freq[j];
		}
		i++;
	}
	memset((void *) u->v, 0, u->n * sizeof(double));	/* zero the freqcounters */
	bstep(u);		/* advance the state */
}

int
main(int argc, char **argv)
{
	struct userdata ud;
	int            *primes, *counters, *hz, *amp;
	int             n = 100;/* number of primes */
	SDL_AudioSpec   desiredSpec;
	SDL_AudioSpec   obtainedSpec;

	/* the primes */
	primes = (int *) malloc(n * sizeof(int));
	init_primes(primes, n);

	/* array of counters modulo primes */
	counters = (int *) malloc(n * sizeof(int));
	memset((void *) counters, 0, n * sizeof(int));

	/* (pseudo-) list of frequencies */
	hz = (int *) malloc(n * sizeof(int));
	memset((void *) hz, 0, n * sizeof(int));

	/* (pseudo-) list of amplitudes */
	amp = (int *) malloc(n * sizeof(int));
	memset((void *) amp, 0, n * sizeof(int));

	/*
	 * first round is meaningless, since all counters are 0. so we
	 * pre-init the frequency list with an end-of-list marker to avoid
	 * output.
	 */
	hz[0] = -1;

	ud.primes = primes;
	ud.counters = counters;
	ud.freq = hz;
	memset(ud.freq, -1, n * sizeof(int));
	ud.amp = amp;
	ud.n = n;
	ud.c = 2;
	ud.v = (double *) malloc(n * sizeof(double));
	memset(ud.v, 0, n * sizeof(double));

	/* Set up the SDL stuff */
	desiredSpec.freq = FREQUENCY;
	desiredSpec.format = AUDIO_S16SYS;
	desiredSpec.channels = 1;
	desiredSpec.samples = 128;
	desiredSpec.callback = audio_callback;
	desiredSpec.userdata = &ud;

	SDL_Init(SDL_INIT_AUDIO);
	SDL_OpenAudio(&desiredSpec, &obtainedSpec);
	SDL_PauseAudio(0);
	while (1)
	{
		SDL_Delay(10);
	}
	SDL_CloseAudio();
}
